# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
Utility module to group commonly used types for type annotation.
"""
from collections.abc import (
    Callable,
    Generator,
    Iterable,
    Sequence,
    Mapping,
)
from typing import (
    Any,
    Union,
)
